local modName = g_currentModName
local modDir = g_currentModDirectory

HusbandryCleaning = {
    SPEC_TABLE_NAME = "spec_" .. modName .. ".husbandryCleaning",
    debug = true,
}

function HusbandryCleaning.prerequisitesPresent(specializations)
    return true
end

function HusbandryCleaning.registerEventListeners(placeableType)
    SpecializationUtil.registerEventListener(placeableType, "onLoad", HusbandryCleaning)
    SpecializationUtil.registerEventListener(placeableType, "onDelete", HusbandryCleaning)
    SpecializationUtil.registerEventListener(placeableType, "onFinalizePlacement", HusbandryCleaning)
    SpecializationUtil.registerEventListener(placeableType, "onPostFinalizePlacement", HusbandryCleaning)
    SpecializationUtil.registerEventListener(placeableType, "onHourChanged", HusbandryCleaning)
end

function HusbandryCleaning.registerXMLPaths(schema, basePath)
    schema:setXMLSpecializationType("Husbandry")
    schema:register(XMLValueType.STRING, basePath .. ".husbandry#cleaningAreasFilename")
    schema:setXMLSpecializationType()
end

function HusbandryCleaning:onLoad(savegame)
    local spec = self[HusbandryCleaning.SPEC_TABLE_NAME]
    local path = nil

    spec.isMod = self.storeItem.isMod
    spec.isDirty = false
    spec.cleaningAreasNode = nil
    spec.cleaningFillType = FillType.MANURE
    spec.cleaningAreas = {}
    spec.cleaningDelta = 0
    spec.animalClusters = {}
    spec.availableAreas = HusbandryCleaning:loadAvailableAreas()
    path = spec.availableAreas[self.i3dFilename]

    if spec.isMod and path == nil then
        local filename = self.xmlFile:getValue("placeable.husbandry#cleaningAreasFilename")

        if filename ~= nil then
            if fileExists(self.baseDirectory .. filename)then
                path = self.baseDirectory .. filename
            end
        end
    end

    if path ~= nil then
        spec.isDirty = true

        local rootNode = self.rootNode
        local linkNode = g_i3DManager:loadSharedI3DFile(path)
        link(rootNode, linkNode)
        spec.cleaningAreasNode = getChildAt(linkNode, 0)
    end
end

function HusbandryCleaning:onDelete()
    if self[HusbandryCleaning.SPEC_TABLE_NAME] ~= nil then
        self[HusbandryCleaning.SPEC_TABLE_NAME] = nil
    end
end

function HusbandryCleaning:onFinalizePlacement()
    local spec = self[HusbandryCleaning.SPEC_TABLE_NAME]
    local cleaningAreasNode = spec.cleaningAreasNode
    local isDirty = spec.isDirty

    if isDirty then
        if getNumOfChildren(cleaningAreasNode) > 0 then
            for i = 1, getNumOfChildren(cleaningAreasNode), 1 do
                local parent = getChildAt(cleaningAreasNode, i - 1)
                local widthNode = getChild(parent, "width")
                local heightNode = getChild(parent, "height")
                local markerNode = getChild(parent, "marker")
                setVisibility(markerNode, true)

                local sx, sy, sz = getWorldTranslation(parent)
                local wx, wy, wz = getWorldTranslation(widthNode)
                local hx, hy, hz = getWorldTranslation(heightNode)
                local lsx, lsy, lsz, lex, ley, lez, radius = DensityMapHeightUtil.getLineByAreaDimensions(sx, sy, sz, wx, wy, wz, hx, hy, hz)

                table.insert(spec.cleaningAreas, {
                    start = parent,
                    width = widthNode,
                    height = heightNode,
                    marker = markerNode,
                    lsx = lsx,
                    lsy = lsy,
                    lsz = lsz,
                    lex = lex,
                    ley = ley,
                    lez = lez,
                    radius = radius
                })
            end
        end
    end
end

function HusbandryCleaning:onPostFinalizePlacement()
    local spec = self[HusbandryCleaning.SPEC_TABLE_NAME]

    for _, area in pairs(spec.cleaningAreas) do
        setVisibility(area.marker, false)
        delete(area.marker)
    end
end

---@param currentHour number
function HusbandryCleaning:onHourChanged(currentHour)
    local spec = self[HusbandryCleaning.SPEC_TABLE_NAME]

    if spec.isDirty == false then
        return
    end

    local fillType = spec.cleaningFillType
    local clusters = self:getClusters()

    if currentHour == 0 then
        local level = HusbandryCleaning:getDirtLevel(spec.cleaningAreas)

        if level > 0 then
            for _, cluster in pairs(clusters) do
                local health = cluster.health

                if health > 10 then
                    cluster:changeHealth(-10)
                end
            end
        end
    end

    if currentHour > 6 and currentHour < 22 then
        -- Animals do sleep, right?
        local litersPerHour = self.spec_husbandryFood.litersPerHour
        local delta = HusbandryCleaning:calculateDelta(clusters, litersPerHour)

        if delta > 0 then
            local minDelta = g_densityMapHeightManager:getMinValidLiterValue(fillType)

            if delta < minDelta then
                delta = minDelta
            end

            local i = math.random(1, #spec.cleaningAreas)
            local cleaningArea = spec.cleaningAreas[i]
            local dropped = HusbandryCleaning:addDirt(cleaningArea, fillType, delta)
        end
    end
end

---@return table
function HusbandryCleaning:loadAvailableAreas(isMod)
    local areas = {}

    if fileExists(modDir .. "xml/cleaningAreas.xml") then
        local xmlFile = XMLFile.load("cleaningAreas", modDir .. "xml/cleaningAreas.xml")

        xmlFile:iterate("cleaningAreas.cleaningArea", function(_, key)
            local match = xmlFile:getString(key .. "#i3dFilename")
            local filename = xmlFile:getString(key .. "#cleaningAreasFilename")
            areas[match] = modDir .. filename
        end)

        xmlFile:delete()
    end

    if isMod then

    end

    return areas
end

---@param area table
---@param fillType number
---@param delta number
---@return number
function HusbandryCleaning:addDirt(area, fillType, delta)
    local foodDropped = 0
    local j = math.random(0, 1)

    if j == 0 then
        local dropped, _ = DensityMapHeightUtil.tipToGroundAroundLine(nil, delta, fillType, area.lsx, area.lsy, area.lsz, area.lex, area.ley, area.lez, 0, area.radius, 0, false, nil, false, true)
        foodDropped = dropped
    else
        local dropped, _ = DensityMapHeightUtil.tipToGroundAroundLine(nil, delta, fillType, area.lex, area.ley, area.lez, area.lsx, area.lsy, area.lsz, 0, area.radius, 0, false, nil, false, true)
        foodDropped = dropped
    end

    return foodDropped
end

---@param animalClusters table
---@param litersPerHour number
---@return number
function HusbandryCleaning:calculateDelta(animalClusters, litersPerHour)
    local delta = 0

    for _, cluster in pairs(animalClusters) do
        local ageFactor = cluster:getAgeFactor() -- Between 0 and 1
        local ageMultiplier = 1 + (ageFactor / 10) -- Between 1 and 1.1

        delta = delta + (litersPerHour * ageMultiplier)
    end

    return math.abs(delta)
end

---@param cleaningAreas table
---@return number
function HusbandryCleaning:getDirtLevel(cleaningAreas, fillType)
    local totalLevel = 0

    for _, cleaningArea in pairs(cleaningAreas) do
        local xs, _, zs = getWorldTranslation(cleaningArea.start)
        local xw, _, zw = getWorldTranslation(cleaningArea.width)
        local xh, _, zh = getWorldTranslation(cleaningArea.height)
        local fillLevel = DensityMapHeightUtil.getFillLevelAtArea(fillType, xs, zs, xw, zw, xh, zh)
        totalLevel = totalLevel + fillLevel
    end

    return totalLevel
end

function HusbandryCleaning:log(message)
    if HusbandryCleaning.debug then
        if type(message) == "table" then
            DebugUtil.printTableRecursively(message, nil, 1, 2)
        else
            Logging.info(message)
        end
    end
end
